#include <PalmOS.h>
#include <VfsMgr.h>
#include "send.h"


#define VFS_SEND_CHUNK_SIZE			128

static const char mMimeTypePrcPdb[] = "application/vnd.palm";
static const char mMimeTypeTextFile[] = "text/plain";
static const char mMimeTypeMP3file[] = "audio/mpeg";
static const char mMimeTypeGeneric[] = "application/octet-stream";
static const char mSendPrefix[] = exgSendBeamPrefix;
static const char mFileExtRecDb[] = ".pdb";
static const char mFileExtResDb[] = ".prc";
static const char mFileExtText[] = ".txt";
static const char mFileExtMP3[] = ".mp3";


static const char* sendPrvMimeTypeFromFilename(const char *filename)
{
	UInt32 len;
	
	len = StrLen(filename);
	if (len >= 4) {
		const char *ext = filename + len - 4;
		
		if (!StrCompare(ext, mFileExtRecDb) || !StrCompare(ext, mFileExtResDb))
			return mMimeTypePrcPdb;
		
		if (!StrCompare(ext, mFileExtText))
			return mMimeTypeTextFile;
		
		if (!StrCompare(ext, mFileExtMP3))
			return mMimeTypeMP3file;
	}
	
	return mMimeTypeGeneric;
}

static Err sendPrvCommonStart(ExgSocketType *sock, const char *mimeType, const char *displayName, const char *fileName, UInt32 approxSize)
{
	MemSet(sock, sizeof(ExgSocketType), 0);
	
	sock->description = (char*)displayName;
	sock->name = (char*)fileName;
	sock->type = (char*)mimeType;
	sock->target = sysFileCLauncher;
	sock->count = 1;
	sock->length = approxSize;
	
	return ExgPut(sock);
}

static Err sendPrvExgDBWriteProc(const void *dataP, UInt32 *sizeP, void *userDataP)
{
	ExgSocketType *sock = (ExgSocketType*)userDataP;
	Err e;
	
	*sizeP = ExgSend(sock, dataP, *sizeP, &e);
	
	return e;
}

static Err sendPrvCommonEnd(ExgSocketType *sock, Err errSoFar)
{
	Err e2;
	
	e2 = ExgDisconnect(sock, errSoFar);
	if (e2 != errNone)
		errSoFar = e2;
	
	return errSoFar;
}

Err sendPrvInternal(UInt16 cardNo, const char *name, UInt32 approxSize)
{
	char *pseudoFileName;
	ExgSocketType sock;
	UInt16 attr;
	Err e;
	
	pseudoFileName = MemPtrNew(StrLen(mSendPrefix) + StrLen(name) + 5);
	if (!pseudoFileName)
		return memErrNotEnoughSpace;
	
	(void)DmDatabaseInfo(cardNo, DmFindDatabase(cardNo, name), NULL, &attr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	
	StrCopy(pseudoFileName, mSendPrefix);
	StrCat(pseudoFileName, name);
	StrCat(pseudoFileName, (attr & dmHdrAttrResDB) ? mFileExtResDb : mFileExtRecDb);
	
	e = sendPrvCommonStart(&sock, mMimeTypePrcPdb, name, pseudoFileName, approxSize);
	if (e != errNone)
		goto out;
	
	e = ExgDBWrite(sendPrvExgDBWriteProc, &sock, name, 0, cardNo);
	
	e = sendPrvCommonEnd(&sock, e);
	
out:
	MemPtrFree(pseudoFileName);
	return e;
}

Err sendPrvFromCard(UInt16 vrn, const char *filename, const char *fullPath)
{
	UInt32 fileSize, done = 0, now, sentDone, sentNow;
	ExgSocketType sock;
	FileRef file;
	UInt8 *buf;
	Err e;
	
	buf = MemPtrNew(VFS_SEND_CHUNK_SIZE);
	if (!buf)
		return memErrNotEnoughSpace;
	
	e = VFSFileOpen(vrn, fullPath, vfsModeRead, &file);
	if (e != errNone)
		goto out;
	
	e = VFSFileSize (file, &fileSize);
	if (e != errNone)
		goto out_closefile;
	
	e = sendPrvCommonStart(&sock, sendPrvMimeTypeFromFilename(filename), filename, filename, fileSize);
	if (e != errNone)
		goto out_closefile;
	
	while (done < fileSize) {
		
		now = fileSize - done;
		if (now > VFS_SEND_CHUNK_SIZE)
			now = VFS_SEND_CHUNK_SIZE;
		
		e = VFSFileRead(file, now, buf, &now);
		if (e == vfsErrFileEOF && now)
			e = errNone;
		if (e != errNone)
			break;
		done += now;
		
		//now send it
		for (sentDone = 0; sentDone < now && e == errNone; sentDone += sentNow) {
		
			sentNow = ExgSend(&sock, buf + sentDone, now - sentDone, &e);
		}
	}
	
	e = sendPrvCommonEnd(&sock, e);
	
out_closefile:
	VFSFileClose(file);

out:
	MemPtrFree(buf);
	return e;
}







