#include <PalmOS.h>


struct NotifSub {
	struct NotifSub *prev, *next;
	SysNotifyProcPtr handlerPtr;
	void *userData;
	UInt16 cardNo;
	LocalID LID;
	Int8 prio;
	Boolean plzDelete;
};

struct NotifSubHead {
	struct NotifSubHead *prev, *next;
	struct NotifSub *subs;
	UInt32 notifType;
	Boolean busy;
};

struct NotifPending {
	struct NotifPending *next;
	SysNotifyParamType notif;
	UInt8 data[];
};

struct Globals {
	struct NotifSubHead *notifSubs;
	struct NotifPending *notifsPendingH, *notifsPendingT;
	UInt8 numIrqBcasts;
};

#define MemPtrNewUnowned(sz)	MemChunkNew(0, sz, memNewChunkFlagNonMovable)


#ifdef __GNUC__

	__attribute__((noinline)) struct Globals* getGlobals(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		
		asm volatile(
			"	movea.l #0xaaaaaaaa, %a0			\n\t"
			"	rts									\n\t"
		);
		return (void*)1;
	}

#else

	asm static struct Globals* getGlobals(void)
	{
		movea.l #0xaaaaaaaa, a0
		rts
	}
	
#endif


static struct NotifSubHead* notifPrvFindHead(struct Globals *g, UInt32 notifType)	//call with semaphore reserved please
{
	struct NotifSubHead *h;
	
	for (h = g->notifSubs; h && h->notifType != notifType; h = h->next);
	
	return h;
}

static struct NotifSub* notifPrvFindSub(struct NotifSubHead *nh, UInt16 cardNo, LocalID LID, Int8 prio)	//call with semaphore reserved please
{
	struct NotifSub *ns;
	
	for (ns = nh->subs; ns; ns = ns->next) {
			
		if (ns->cardNo != cardNo)
			continue;
		
		if (ns->LID != LID)
			continue;
		
		if (ns->prio != prio)
			continue;
		
		return ns;
	}
	
	return NULL;
}

static Err pSysNotifyRegister(UInt16 cardNo, LocalID LID, UInt32 notifType, SysNotifyProcPtr callbackP, Int8 prio, void *userDataP)
{
	struct NotifSub *ns, *p, *cur;
	struct NotifSubHead *nh;
	struct Globals *g;
	Err e = errNone;
	
	if (LID == sysNotifyNoDatabaseID && !callbackP)
		return sysErrParamErr;
	
	g = getGlobals();
	
	MemSemaphoreReserve(false);
	nh = notifPrvFindHead(g, notifType);
	if (!nh) {
		nh = MemPtrNewUnowned(sizeof(struct NotifSubHead));
		if (!nh)
			e = memErrNotEnoughSpace;
		else {
			nh->busy = false;
			nh->notifType = notifType;
			nh->subs = NULL;
			nh->prev = NULL;
			nh->next = g->notifSubs;
			g->notifSubs = nh;
			if (nh->next)
				nh->next->prev = nh;
		}
	}
	
	if (e == errNone && notifPrvFindSub(nh, cardNo, LID, prio))
		e = sysNotifyErrDuplicateEntry;
	
	if (e == errNone) {
		ns = MemPtrNewUnowned(sizeof(struct NotifSub));
		if (!ns)
			e = memErrNotEnoughSpace;
		else {
			ns->handlerPtr = callbackP;
			ns->userData = userDataP;
			ns->cardNo = cardNo;
			ns->LID = LID;
			ns->prio = prio;
			ns->plzDelete = false;
			
			for (p = NULL, cur = nh->subs; cur && cur->prio < prio; p = cur, cur = cur->next);
			if (p) {
				ns->prev = p;
				ns->next = p->next;
				p->next = ns;
			}
			else {
				ns->prev = NULL;
				ns->next = nh->subs;
				nh->subs = ns;
			}
			if (ns->next)
				ns->next->prev = ns;
		}
	}
	
	MemSemaphoreRelease(false);
	return e;
}

static Boolean notifPrvDeleteSub(struct Globals *g, struct NotifSubHead* nh, struct NotifSub *ns)	//return true if notif sub head went away
{
	Boolean ret = false;
	
	if (nh->busy)
		ns->plzDelete = true;
	else {
		if (ns->next)
			ns->next->prev = ns->prev;
		if (ns->prev)
			ns->prev->next = ns->next;
		else
			nh->subs = ns->next;
		MemPtrFree(ns);
		
		if (!nh->subs) {
			
			if (nh->next)
				nh->next->prev = nh->prev;
			if (nh->prev)
				nh->prev->next = nh->next;
			else
				g->notifSubs = nh->next;
			
			MemPtrFree(nh);
			ret = true;
		}
	}
	
	return ret;
}

static Err pSysNotifyUnregister(UInt16 cardNo, LocalID LID, UInt32 notifType, Int8 prio)
{
	struct Globals *g = getGlobals();
	struct NotifSubHead *nh;
	struct NotifSub *ns;
	Err e = errNone;
	
	MemSemaphoreReserve(false);
	nh = notifPrvFindHead(g, notifType);
	if (!nh)
		e = sysNotifyErrEntryNotFound;
	else {
		ns = notifPrvFindSub(nh, cardNo, LID, prio);
		if (!ns)
			e = sysNotifyErrEntryNotFound;
		else
			notifPrvDeleteSub(g, nh, ns);
	}
	MemSemaphoreRelease(false);
	return e;
}

static void notifPrvDbDeleted(struct Globals *g, UInt16 cardNo, LocalID LID)
{
	struct NotifSubHead *nh;
	struct NotifSub *ns;
	
	MemSemaphoreReserve(false);

		
restart:
	for (nh = g->notifSubs; nh; nh = nh->next) {
		for (ns = nh->subs; ns; ns = ns->next) {
			if (ns->cardNo == cardNo && ns->LID == LID && !ns->plzDelete) {
				
				notifPrvDeleteSub(g, nh, ns);
				goto restart;
			}
		}
	}
	
	MemSemaphoreRelease(false);
}

static Err notifPrvSend(struct Globals *g, SysNotifyParamType *notifP)	//call with semaphore held plz
{
	struct NotifSubHead *nh = notifPrvFindHead(g, notifP->notifyType);
	SysNotifyParamType notif;
	struct NotifSub *ns;
	Err e = errNone;
	
	MemMove(&notif, notifP, sizeof(notif));

	if (nh) {
		
		Boolean wasBusy = nh->busy;
		nh->busy = true;	//prevent deletions from messing things up. insertions we'll ignore and either or not send things to them
		
		for (ns = nh->subs; ns; ns = ns->next) {
			
			if (ns->plzDelete)
				continue;
			
			notif.userDataP = ns->userData;
			
			MemSemaphoreRelease(false);
			
			if (ns->handlerPtr)
				e = ns->handlerPtr(&notif);
			else {
				UInt32 appRet;
				
				e = SysAppLaunch(ns->cardNo, ns->LID, 0, sysAppLaunchCmdNotify, &notif, &appRet);
				if (e == errNone)
					e = (Err)appRet;
			}

			MemSemaphoreReserve(false);
			
			if (e == sysNotifyErrBroadcastCancelled)	//do not ask, but system does this too
				break;
		}
		
		nh->busy = wasBusy;
		
		//now delete all that was deleted. this is slow slow, but i also do not care
		if (!wasBusy) while(1) {
			for (ns = nh->subs; ns && !ns->plzDelete; ns = ns->next);
			if (!ns)
				break;
			if (notifPrvDeleteSub(g, nh, ns))
				break;
		}
	}
	
	return errNone;
}

static Err pSysNotifyBroadcast(SysNotifyParamType *notif)
{
	struct Globals *g = getGlobals();
	Err e;
	
	MemSemaphoreReserve(false);
	e = notifPrvSend(g, notif);
	MemSemaphoreRelease(false);
	
	return e;
}

static Err pSysNotifyBroadcastDeferred(SysNotifyParamType *notif, UInt16 paramSize)
{
	struct NotifPending *np;
	struct Globals *g;
	
	np = MemPtrNewUnowned(sizeof(struct NotifPending) + paramSize);
	if (!np)
		return memErrNotEnoughSpace;
	
	np->next = NULL;
	MemMove(&np->notif, notif, sizeof(np->notif));
	MemMove(np->data, notif->notifyDetailsP, paramSize);
	if (paramSize)
		np->notif.notifyDetailsP = np->data;
	
	g = getGlobals();
	MemSemaphoreReserve(false);
	if (g->notifsPendingT) 
		g->notifsPendingT->next = np;
	else
		g->notifsPendingH = np;
	g->notifsPendingT = np;
	MemSemaphoreRelease(false);
	
	EvtWakeup();
	
	return errNone;
}

static void notifPrvSendAllPending(struct Globals *g)
{
	struct NotifPending *np;
	Boolean gotOne;
	
	MemSemaphoreReserve(false);
	
	while ((np = g->notifsPendingH)) {
		
		g->notifsPendingH = np->next;
		if (!g->notifsPendingH)
			g->notifsPendingT = NULL;
		
		notifPrvSend(g, &np->notif);
		MemPtrFree(np);
	}

	MemSemaphoreRelease(false);
}

Err pDmDeleteDatabase(Err (*oldTrap)(UInt16 cardNo, LocalID LID), void* retAddr, UInt16 cardNo, LocalID LID)
{
	Err e;
	
	e = oldTrap(cardNo, LID);
	if (e == errNone)
		notifPrvDbDeleted(getGlobals(), cardNo, LID);
	
	return e;
}

void pEvtGetEvent(void (*oldTrap)(EventType *event, Int32 timeout), void *retAddr, EventType *event, Int32 timeout)
{
	struct Globals *g = getGlobals();
	
	notifPrvSendAllPending(g);
	oldTrap(event, timeout);
	notifPrvSendAllPending(g);
}

static const struct SysNotifyParamType mResetDoneNotif = {.notifyType = sysNotifyResetFinishedEvent, .broadcaster = sysNotifyBroadcasterCode, };



#ifdef __GNUC__

	static void pEvtGetEventWrapper(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		//also gcc gets too clever with tailcalls, so to avoid them, we provide a proper stack
		//sadly this means we need to create pointless stack frames
		//c'est la vie
		
		asm volatile(
			"	move.l  #0xaaaaaaaa, %d0			\n\t"		//old trap
			"	move.l  8(%sp), -(%sp)				\n\t"
			"	move.l  8(%sp), -(%sp)				\n\t"
			"	clr.l   -(%sp)						\n\t"
			"	move.l  %d0, -(%sp)					\n\t"
			"	bsr    pEvtGetEvent					\n\t"
			"	lea    16(%sp), %sp					\n\t"
		);
	}

	static void pDmDeleteDatabaseWrapper(void)
	{
		//with no "naked" attribute available, this is a bit fragile, but as long as you do not try to compile with -O0, it will be ok
		//also gcc gets too clever with tailcalls, so to avoid them, we provide a proper stack
		//sadly this means we need to create pointless stack frames
		//c'est la vie
		
		asm volatile(
			"	move.l  #0xaaaaaaaa, %d0			\n\t"		//old trap
			"	move.l  6(%sp), -(%sp)				\n\t"
			"	move.w  8(%sp), -(%sp)				\n\t"
			"	clr.l   -(%sp)						\n\t"
			"	move.l  %d0, -(%sp)					\n\t"
			"	bsr    pDmDeleteDatabase			\n\t"
			"	lea    14(%sp), %sp					\n\t"
		);
	}


#else

	asm void pEvtGetEventWrapper(void)
	{
		move.l #0xaaaaaaaa, -(sp)			//old trap
		jsr    pEvtGetEvent
		lea    4(sp), sp
		rts
	}
	
	asm void pDmDeleteDatabaseWrapper(void)
	{
		move.l #0xaaaaaaaa, -(sp)			//old trap
		jsr    pDmDeleteDatabase
		lea    4(sp), sp
		rts
	}

#endif


void __Startup__(void)
{
	UInt32 dummy;
	
	if (errNone != FtrGet(sysFtrCreator, sysFtrNumNotifyMgrVersion, &dummy) || !dummy) {
				
		struct Globals *g = MemPtrNewUnowned(sizeof(struct Globals));
		
		MemSet(g, sizeof(*g), 0);
		
		MemSemaphoreReserve(true);
		*(UInt32*)(((UInt16*)getGlobals) + 1) = (UInt32)g;
		*(UInt32*)(((UInt16*)pEvtGetEventWrapper) + 1) = (UInt32)SysGetTrapAddress(sysTrapEvtGetEvent);
		*(UInt32*)(((UInt16*)pDmDeleteDatabaseWrapper) + 1) = (UInt32)SysGetTrapAddress(sysTrapDmDeleteDatabase);
		MemSemaphoreRelease(true);
		
		SysSetTrapAddress(sysTrapSysNotifyRegister, pSysNotifyRegister);
		SysSetTrapAddress(sysTrapSysNotifyUnregister,pSysNotifyUnregister);
		SysSetTrapAddress(sysTrapSysNotifyBroadcast, pSysNotifyBroadcast);
		SysSetTrapAddress(sysTrapSysNotifyBroadcastDeferred, pSysNotifyBroadcastDeferred);
		SysSetTrapAddress(sysTrapEvtGetEvent, pEvtGetEventWrapper);
		SysSetTrapAddress(sysTrapDmDeleteDatabase, pDmDeleteDatabaseWrapper);
		
		FtrSet(sysFtrCreator, sysFtrNumNotifyMgrVersion, 1);
		
		SysNotifyBroadcastDeferred((struct SysNotifyParamType*)&mResetDoneNotif, 0);
		
		MemHandleLock(DmGet1Resource('extn', 12));
	}
}
