#include <SerialMgrOld.h>
#include <PalmOS.h>
#include "bbUart.h"
#include "update.h"


#define UPDATE_BAUD_RATE	19200
#define SECTOR_SIZE			128
#define UPDATE_STAGE		0x10000600ul


#define DELAY_ONE_SEC		0x80000
#define DELAY_100_MS		0xCCCC
#define PROPER_STOP_BIT		0x200

//expected stuff MUST be a null-termed constant string
#define SEND_AND_RECV(_tx, _txLen, _expect, _echo)																				\
	do { 																														\
		e = updatePrvSendAndRecv(state, _tx, _txLen, _buf, sizeof(_expect) - 1, &nRxed, DELAY_ONE_SEC /* one second */, _echo);	\
		if (e != errNone)																										\
			goto out_close_port;																								\
		if (nRxed != sizeof(_expect) - 1 || !MemIsEqual(_buf, _expect, sizeof(_expect) - 1)) {									\
			e = serErrLineErr;																									\
			goto out_close_port;																								\
		}																														\
	} while(0)

#define SEND_AND_RECV_CONST_WITH_ECHO(_tx, _expect)		SEND_AND_RECV(_tx, sizeof(_tx) - 1, _expect, true)
#define SEND_AND_RECV_CONST_NO_ECHO(_tx, _expect)		SEND_AND_RECV(_tx, sizeof(_tx) - 1, _expect, false)


//MemCmp is PalmOS 2.0+
static Boolean MemIsEqual(const void *aP, const void *bP, UInt16 len)
{
	const UInt8* a = aP;
	const UInt8* b = bP;
	
	while (len--) {
		
		if (*a++ != *b++)
			return false;
	}
	return true;
}

//complex since we need to RX right after TX and need to be ready for it!
static Err updatePrvSendAndRecv(struct UartState *state, const void *txP, UInt16 nSend, void *rx, UInt16 maxRx, UInt16 *nRecvd, UInt32 maxWait, Boolean expectEcho)
{
	const UInt8 *tx = (const UInt8*)txP;
	UInt8 *rxTmp, *rxStart, rxEcho;
	UInt16 now;
	
	//our send provides no "stop" bit to allow it to rx faster
	//if we intend to not rx, we need to send a stop bit
	//this is done by "waiting to rx" for a small period of time
	//after sending a byte
	
	while (nSend > 1) {
		
		now = bbUArtTxThenRx(state, (UInt16)*tx++, &rxEcho, 1, expectEcho ? maxWait : PROPER_STOP_BIT);
		if (expectEcho) {
			
			if (now != 1 || rxEcho != tx[-1])
				return serErrLineErr;
		}
		nSend--;
	}
	
	if (nSend) {	//send last char
		
		rxTmp = MemPtrNew(1 + maxRx);	//if this fails, let it crash :)
		
		now = bbUArtTxThenRx(state, (UInt16)*tx, rxTmp, expectEcho ? (maxRx + 1) : (maxRx ? maxRx : 1), (expectEcho || maxRx) ? maxWait : PROPER_STOP_BIT);
		if (!expectEcho)
			rxStart = rxTmp;
		else {
			
			if (now < 1 || rxTmp[0] != tx[0]) {
				MemPtrFree(rxTmp);
				return serErrLineErr;
			}
			rxStart = rxTmp + 1;
			now--;
		}
		
		if (rx)
			MemMove(rx, rxStart, now);
		MemPtrFree(rxTmp);
	}
	else {			//no chars to send at all - just rx
		
		now = bbUArtTxThenRx(state, -1, rx, maxRx, maxWait);
	}
	
	*nRecvd = now;
	
	return errNone;
}


static Err updateSendBlock(struct UartState *state, char *_buf, const UInt8 *data, UInt32 addrTo, UInt16 numUsefulBytes)
{
	UInt16 nRxed, numPaddingBytes = SECTOR_SIZE - numUsefulBytes;
	UInt8 lastByte;
	Err e;
	
	
	//prepare all sectors - makes life easier
	SEND_AND_RECV_CONST_NO_ECHO("P 0 15\r\n", "0\r\n");
	
	//send request to upload (do not use StrPrintF since PalmOS 1.0 lacks it)
	_buf[0] = 'W';
	_buf[1] = ' ';
	StrIToA(_buf + 2, UPDATE_STAGE);
	StrCat(_buf, " ");
	StrIToA(_buf + StrLen(_buf), SECTOR_SIZE);
	StrCat(_buf,"\r\n");
	SEND_AND_RECV(_buf, StrLen(_buf), "0\r\n", false);
	
	//docs say that after the data upload, an "OK\r\n" message is sent, but in reality no such thing happens and firmware analysis proves it never will be

	//send data 
	e = updatePrvSendAndRecv(state, data, numUsefulBytes, NULL, 0, &nRxed, PROPER_STOP_BIT, false);
	if (e != errNone)
		goto out_close_port;
	
	//send padding
	while (numPaddingBytes--) {
		
		UInt8 zero = 0;
		
		e = updatePrvSendAndRecv(state, &zero, 1, NULL, 0, &nRxed, PROPER_STOP_BIT, false);
		if (e != errNone)
			goto out_close_port;
	}
	
	//send write command
	_buf[0] = 'C';
	_buf[1] = ' ';
	StrIToA(_buf + 2, addrTo);
	StrCat(_buf, " ");
	StrIToA(_buf + StrLen(_buf), UPDATE_STAGE);
	StrCat(_buf, " ");
	StrIToA(_buf + StrLen(_buf), SECTOR_SIZE);
	StrCat(_buf,"\r\n");
	SEND_AND_RECV(_buf, StrLen(_buf), "0\r\n", false);


out_close_port:
	return e;
}

Err updatePerform(const void *dataP, UInt32 size, void (*UpdateProgressF)(void *userData, UInt16 blocksDone, UInt16 blocksOf), void *userData)
{
	UInt16 i, nRxed, nBlocks = (size + SECTOR_SIZE - 1) / SECTOR_SIZE;
	struct UartState us, *state = &us;
	enum UartOption uo;
	const UInt8* src;
	char _buf[64];
	Err e;
	
	if (!nBlocks)
		return sysErrParamErr;
	
	uo = portPrvFind();
	if (uo == UartUnknown) {
		e = serErrNoDevicesAvail;
		goto out;
	}
	
	if (!bbUartOpen(&us, uo)) {
		e = serErrNotOpen;
		goto out;
	}
	
	UpdateProgressF(userData, 0, nBlocks);	//tell UI we're starting
	
	//we sync a a special way since LPCuart starts weirdly
	(void)updatePrvSendAndRecv(state, "?", 1, NULL, 0, &nRxed, PROPER_STOP_BIT, false);
	SysTaskDelay(100);
	
	//sync
	SEND_AND_RECV_CONST_WITH_ECHO("Synchronized\r\n", "OK\r\n");
	SEND_AND_RECV_CONST_WITH_ECHO("12000\r\n", "OK\r\n");
	
	//echo off
	SEND_AND_RECV_CONST_WITH_ECHO("A 0\r\n", "0\r\n");
	
	//unlock
	SEND_AND_RECV_CONST_NO_ECHO("U 23130\r\n", "0\r\n");
	
	//erase
	SEND_AND_RECV_CONST_NO_ECHO("P 0 15\r\n", "0\r\n");
	SEND_AND_RECV_CONST_NO_ECHO("E 0 15\r\n", "0\r\n");
	
	//write all but first
	for (i = 1, src = ((UInt8*)dataP) + SECTOR_SIZE; i < nBlocks; i++, src += SECTOR_SIZE) {
		
		
		e = updateSendBlock(state, _buf, src, src - (UInt8*)dataP, (i == nBlocks - 1) ? (size % SECTOR_SIZE) : SECTOR_SIZE);
		if (e != errNone)
			goto out_close_port;
		
		UpdateProgressF(userData, i, nBlocks);
	}
	
	//write first
	updateSendBlock(state, _buf, dataP, 0, (i == nBlocks - 1) ? (size % SECTOR_SIZE) : SECTOR_SIZE);
	if (e != errNone)
		goto out_close_port;
	
	UpdateProgressF(userData, nBlocks, nBlocks);


	//go
	SEND_AND_RECV_CONST_NO_ECHO("G 0 T\r\n", "0");	//we only get a zero...
	

out_close_port:
	bbUartClose(&us);

out:
	return e;
}