#include <stdio.h>
#include <alloca.h>
#include <unistd.h>
#include <string.h>
#include "printf.h"
#include "hotsync.h"
#include "swap.h"
#include "frag.h"
#include "slk.h"


static struct FragSendState mFs;
static uint8_t mTxTid;

//high level packets (replies have 0x80 bit set)
#define HS_PKT_REPLY_MASK		0x80

#define HS_PKT_WAKE				0x01
#define HS_PKT_INIT				0x02
#define HS_PKT_GET_USER_INFO	0x10	//data is struct HsEmptyRequest
#define HS_PKT_GET_SYS_INFO		0x12	//required, else newer devices refuse to talk to us. param is struct HsHotsyncVersion
#define HS_PKT_GET_STORE_INFO	0x15	//data in is HsStoreInfoReq, reply has HsStoreInfoResp as arg1, OS2+ sends more in arg2 as well but whatever
#define HS_PKT_GET_DB_LIST		0x16
#define HS_PKT_DB_OPEN			0x17	//data is struct HsDbOpenOrDeleteReq 
#define HS_PKT_DB_CREATE		0x18	//data is struct HsDbCreateReq
#define HS_PKT_DB_CLOSE			0x19	//data is struct HsCloseDbReq
#define HS_PKT_DB_DELETE		0x1a	//data is struct HsDbOpenOrDeleteReq
#define HS_PKT_APP_INFO_READ	0x1b	//data is struct HsInfoBlockReadReq
#define HS_PKT_APP_INFO_WRITE	0x1c	//data is struct HsInfoBlockWriteReq
#define HS_PKT_SORT_INFO_READ	0x1d	//data is struct HsInfoBlockReadReq
#define HS_PKT_SORT_INFO_WRITE	0x1e	//data is struct HsInfoBlockWriteReq
#define HS_PKT_REC_READ			0x20	//data is struct HsRecReadByIdxReq	//and it is param 0x21!!!
#define HS_PKT_REC_WRITE		0x21	//data is struct HsRecWriteReq
#define HW_PKT_REC_DELETE		0x22	//data is struct HsRecDeleteReq
#define HS_PKT_RES_READ			0x23	//data is struct HsResReadByIdxReq
#define HS_PKT_RES_WRITE		0x24	//data is struct HsResWriteReq
#define HW_PKT_RES_DELETE		0x25	//data is struct HsResDeleteReq
#define HS_PKT_REQ_RESET		0x29	//no data
#define HS_PKT_RPC				0x2d	//data in/out is struct HsRpcPkt, no wrapping 
#define HS_PKT_OPEN_CONDUIT		0x2e	//data is struct HsEmptyRequest
#define HS_PKT_SYNC_END			0x2f	//data is struct HsSyncEndReq

#define HS_ERR_SYSTEM			1
#define HS_ERR_UNKNOWN_REQ		2
#define HS_ERR_OOM				3
#define HS_ERR_PARAM_ERR		4
#define HS_ERR_NOT_FOUND		5


struct HsUnalignedU32 {
	uint32_t val;
} __attribute__((packed));

struct HsUnalignedU16 {
	uint16_t val;
} __attribute__((packed));

//wakup (from pilot) & init (from us) packets follow this structure, others use a higher level yet...
struct HsInitPkt {
	uint8_t pktTyp;
	uint8_t flags;			//HS_INIT_FLG_CHANGE_BAUDRATE to change baud rate
	uint8_t ver[2];
	uint16_t sbz;
	uint32_t maxBaud;		//BE, actual baudrate on reply if HS_INIT_FLG_CHANGE_BAUDRATE is set	
} __attribute__((packed));

#define HS_INIT_FLG_CHANGE_BAUDRATE		0x80

//this is the higher level
struct HsHlReqHdr {	//must be 4 bytes or larger
	uint8_t pktTyp;
	uint8_t nArgs;
} __attribute__((packed));

struct HsHlRspHdr {	//must be 6(!?) bytes or larger
	uint8_t pktTyp;
	uint8_t nArgs;
	uint16_t err;
} __attribute__((packed));

//each arg is wrapped in a thing that specifies its length and all args seem to need to start on an even byte boundary
//each arg is named with a number starting at 0x20. this is in the argumentID byte (first byte), the top 2 bits indicate size type:
//	0x80 set means it is a "small arg" (16-bit length)
//		then there is a padding byte, and then the 16-bit arg length
//	0x40 set means "long arg" (32-bit length)
//		then there is a padding byte, and then the 32-bit arg length
//	else it is a tiny arg
//		then there is a length
//	long args seems to not always be supported and we do not use them
//technically it is possiblt to include arg #0x21 without ever sending arg 0x20, but we do not support that as we never see that used

#define HS_HL_ARG_MASK_SMALL	0x80
#define HS_HL_ARG_MASK_LONG		0x40
#define HS_HL_ARG_MASK_IDX		0x3f
#define HS_HL_ARG_IDX_FIRST		0x20

#define HS_END_CODE_SUCCESS		0x0000
#define HS_END_CODE_OTHER		0x0003

struct HsHlArgWrapperTiny {
	uint8_t argIdAndLenType;
	uint8_t argLen;
};

struct HsHlArgWrapperSmall {
	uint8_t argIdAndLenType;
	uint8_t rfu;
	uint16_t argLen;
};

struct HsEmptyRequest {
	uint16_t sbz;
} __attribute__((packed));

struct HsSyncEndReq {
	uint16_t endCode;
} __attribute__((packed));


struct HsDbEnumResult {
	uint16_t lastIndex;				//index of last entry RXed (we only get one at a time)
	uint8_t enumFlags;				//top bit means thee are more...
	uint8_t numItems;				//should always be one
	struct HsDbInfo dbi;
} __attribute__((packed));


struct HsHotsyncVersion {		//if we claim to be under ver 1.2, some newer devices will not talk to us...
	uint16_t major;
	uint16_t minor;	
} __attribute__((packed));

struct HsSysInfoResp {
	uint32_t romVer;
	uint32_t locVer;
	uint8_t rfu;
	uint8_t prodIdLen;
	uint8_t prodId[];
} __attribute__((packed));

struct HsCloseDbReq {
	uint8_t dbHandle;
} __attribute__((packed));

struct HsDbCreateRsp {
	uint8_t dbHandle;
} __attribute__((packed));

struct HsEnumDbsReq {
	uint8_t flags;		//0x80 for ram, 0x40 for rom
	uint8_t cardNo;
	uint16_t startIdx;
} __attribute__((packed));

struct HsDbOpenOrDeleteReq {
	uint8_t cardNo;
	uint8_t openMode;
	char dbName[32];
} __attribute__((packed));

struct HsInfoBlockReadReq {
	uint8_t dbHandle;
	uint8_t rfu;
	uint16_t ofst;
	uint16_t len;
} __attribute__((packed));

struct HsInfoBlockReadResp {
	uint16_t fullSize;
	uint8_t data[];
} __attribute__((packed));

struct HsInfoBlockWriteReq {
	uint8_t dbHandle;
	uint8_t rfu;
	uint16_t size;
	uint8_t data[];
} __attribute__((packed));

struct HsRecReadByIdxReq {
	uint8_t dbHandle;
	uint8_t rfu;
	uint16_t idx;
	uint16_t ofst;
	uint16_t len;
} __attribute__((packed));

struct HsRecReadResp {
	uint32_t recId;
	uint16_t recIdx;
	uint16_t recSize;
	uint8_t attr;
	uint8_t category;
	uint8_t data[];
} __attribute__((packed));

struct HsRecWriteReq {
	uint8_t dbHandle;
	uint8_t flags;			//must be 0x80
	uint32_t recID;
	uint8_t recAttrs;
	uint8_t category;
	uint8_t data[];
} __attribute__((packed));

struct HsRecWriteResp {
	uint32_t recID;
} __attribute__((packed));

struct HsRecDeleteReq {
	uint8_t dbHandle;
	uint8_t flags;			//zero for our use cases
	uint32_t recID;
} __attribute__((packed));

struct HsResWriteReq {
	uint8_t dbHandle;
	uint8_t rfu;
	uint32_t resTyp;
	uint16_t resID;
	uint16_t resSize;
} __attribute__((packed));

struct HsResReadByIdxReq {
	uint8_t dbHandle;
	uint8_t rfu;
	uint16_t idx;
	uint16_t ofst;
	uint16_t len;
} __attribute__((packed));

struct HsResReadResp {
	uint32_t resType;
	uint16_t resID;
	uint16_t idx;
	uint16_t resSize;
	uint8_t data[];
} __attribute__((packed));

struct HsResDeleteReq {
	uint8_t dbHandle;
	uint8_t flags;			//zero for our use cases
	uint32_t resType;
	uint16_t resID;
} __attribute__((packed));

struct HsStoreInfoReq {
	uint8_t cardNo;
	uint8_t rfu;
} __attribute__((packed));

struct HsStoreInfoResp {
	uint8_t lastCard;					//index of last card in the returned array
	char haveMore;						//boolean if there are more cards in the device
	uint8_t rfu;
	uint8_t numCards;					//how many were returned in this reply
	
	//struct HsCardInfo[numCards]
	
} __attribute__((packed));









static uint8_t hotsyncPrvGetTxTid(void)
{
	mTxTid++;
	if (mTxTid == 0xff)
		mTxTid = 1;
	
	return mTxTid;
}

static bool hotsyncPrvTx(const void *src, uint16_t len)
{
	struct SlkHdr hdr = {.dstSlk = PKT_ADDR_DESKTOP_LINK, .srcSck = PKT_ADDR_DESKTOP_LINK, .pktTyp = PKT_TYP_HOTSYNC, .tid = hotsyncPrvGetTxTid()};
	
	return fragSend(NULL, &hdr, src, len, NULL);
}

static bool hotsyncPrvRx(void *dst, uint16_t *lenP)
{
	uint32_t endTime = slkExtGetCurTime() + 100;	//wait at most a second
	struct SlkHdr rxSlkHdr;
	uint16_t rxLen;
	
	do {
		
		rxLen = *lenP;
		
		if (!fragRecv(NULL, &rxSlkHdr, dst, &rxLen, endTime, true, NULL))
			return false;
		
	} while (rxSlkHdr.dstSlk != PKT_ADDR_DESKTOP_LINK || rxSlkHdr.srcSck != PKT_ADDR_DESKTOP_LINK || rxSlkHdr.pktTyp != PKT_TYP_HOTSYNC);
	
	*lenP = rxLen;
	
	return true;
}

//parse arg wrappers. returns number of args
static int_fast8_t hotsyncPrvArgsParse(const void *pktBufP, uint16_t rxLen, uint8_t maxArgs, uint16_t *argOfstP)	//array shoudl be [maxArgs + 1]
{	
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)pktBufP;
	const uint8_t *pktBuf = (const uint8_t*)pktBufP, *ptr = (const uint8_t*)(rsp + 1), *end = pktBuf + rxLen;
	uint_fast8_t i;
	
	if (rsp->nArgs > maxArgs)
		return 0;
	
	for (i = 0; i < rsp->nArgs; i++) {
		
		const struct HsHlArgWrapperSmall *small = (const struct HsHlArgWrapperSmall*)ptr;
		const struct HsHlArgWrapperTiny *tiny = (const struct HsHlArgWrapperTiny*)ptr;
		uint_fast16_t remain = end - ptr, len;
		
		if ((tiny->argIdAndLenType & HS_HL_ARG_MASK_IDX) != HS_HL_ARG_IDX_FIRST + i) {
			
			//fprintf(stderr, "arg idx %02x invalid\n", tiny->argIdAndLenType);
			return -1;
		}
		
		if ((tiny->argIdAndLenType & HS_HL_ARG_MASK_SMALL) && remain >= sizeof(struct HsHlArgWrapperSmall)) {
			
			ptr += sizeof(struct HsHlArgWrapperSmall);
			remain -= sizeof(struct HsHlArgWrapperSmall);
			
			len = TO_U16(small->argLen);
		}
		else if (!(tiny->argIdAndLenType & HS_HL_ARG_MASK_LONG) && remain >= sizeof(struct HsHlArgWrapperTiny)) {
			
			ptr += sizeof(struct HsHlArgWrapperTiny);
			remain -= sizeof(struct HsHlArgWrapperTiny);
			
			len = tiny->argLen;
		}
		else {
			
			//fprintf(stderr, "invalid packet param parsing result\n");
			return -1;
		}
		
		if (len > remain) {
			
			//fprintf(stderr, "arg ends past packet\n");
			return 0;
		}
		argOfstP[i] = ptr - pktBuf;
		ptr += len;
		
		//fprintf(stderr, "arg %u is @ %u bytes\n", i, argOfstP[i]);
	}
	if (ptr != end) {
		
		//fprintf(stderr, "bytes left over after param parsing\n");
		return -1;
	}
	argOfstP[i] = ptr - pktBuf;
	//fprintf(stderr, "arg end is @ %u bytes\n", argOfstP[i]);
	
	return i;
}

uint16_t hotsyncGetUserInfo(struct HsUserInfo *info, uint16_t memLen)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsEmptyRequest uir;
		
	} __attribute__((packed)) static const userInfoReq = {
		.hdr = {
			.pktTyp = HS_PKT_GET_USER_INFO,
			.nArgs = 0,
		},
		.uir = {
			.sbz = 0,
		},
	};
	
	if (!hotsyncPrvTx(&userInfoReq, sizeof(userInfoReq)))
		return 0;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return 0;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_GET_USER_INFO) || rsp->err)
		return 0;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsUserInfo))
		return 0;
	
	if (memLen > argsOfst[1] - argsOfst[0])
		memLen = argsOfst[1] - argsOfst[0];
	
	memcpy(info, mTxRxBuf + argsOfst[0], memLen);
	
	return memLen;
}

bool hotsyncGetCardInfo(uint8_t cardNo, HsCardInfoCbk cbkF, void *cbkD)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	const struct HsStoreInfoResp *cih;
	const struct HsCardInfo *ci;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsStoreInfoReq sir;
	} __attribute__((packed)) storeInfoReq = {
		.hdr = {
			.pktTyp = HS_PKT_GET_STORE_INFO,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsStoreInfoReq),
		},
		.sir = {
			.cardNo = cardNo,
		},
	};
	
	if (!hotsyncPrvTx(&storeInfoReq, sizeof(storeInfoReq)))
		return false;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return false;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_GET_STORE_INFO) || rsp->err)
		return false;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsStoreInfoResp))
		return false;
	
	cih = (const struct HsStoreInfoResp*)(mTxRxBuf + argsOfst[0]);
	if (!cih->numCards)
		return false;
		
	if (argsOfst[1] - argsOfst[0] < sizeof(struct HsStoreInfoResp) + sizeof(struct HsCardInfo))
		return false;
	ci = (const struct HsCardInfo*)(cih + 1);
	
	if (argsOfst[1] - argsOfst[0] < sizeof(struct HsStoreInfoResp) + ci->structSz)
		return false;
	
	cbkF(cbkD, (const struct HsCardInfo*)(cih + 1));
	
	return true;
}

uint16_t hotsyncGetSysInfo(struct HsSysInfo *info, uint16_t memLen)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsHotsyncVersion hsv;
	} __attribute__((packed)) static const sysInfoReq = {
		.hdr = {
			.pktTyp = HS_PKT_GET_SYS_INFO,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsHotsyncVersion),
		},
		.hsv = {
			.major = TO_U16(1),		//we are v1.2
			.minor = TO_U16(2),
		},
	};
	
	if (!hotsyncPrvTx(&sysInfoReq, sizeof(sysInfoReq)))
		return 0;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return 0;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_GET_SYS_INFO) || rsp->err)
		return 0;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsSysInfo))
		return 0;
	
	if (memLen > argsOfst[1] - argsOfst[0])
		memLen = argsOfst[1] - argsOfst[0];
	
	memcpy(info, mTxRxBuf + argsOfst[0], memLen);
	
	return memLen;
}

bool hotsyncOpenConduit(void)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsEmptyRequest uir;
		
	} __attribute__((packed)) static const openConduitReq = {
		.hdr = {
			.pktTyp = HS_PKT_OPEN_CONDUIT,
			.nArgs = 0,
		},
	};
	
	if (!hotsyncPrvTx(&openConduitReq, sizeof(openConduitReq)))
		return 0;

	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return 0;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_OPEN_CONDUIT) || rsp->err)
		return 0;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	
	return nArgs >= 0;
}

bool hotsyncDemandResetAtEnd(void)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsEmptyRequest uir;
		
	} __attribute__((packed)) static const resetReq = {
		.hdr = {
			.pktTyp = HS_PKT_REQ_RESET,
			.nArgs = 0,
		},
		.uir = {
			.sbz = 0,
		},
	};
	
	if (!hotsyncPrvTx(&resetReq, sizeof(resetReq)))
		return 0;

	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return 0;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_REQ_RESET) || rsp->err)
		return 0;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	
	return nArgs >= 0;
}

bool hotsyncEndSync(enum HsSyncEndCause cause)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsSyncEndReq er;
	} __attribute__((packed)) endReq = {
		.hdr = {
			.pktTyp = HS_PKT_SYNC_END,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsSyncEndReq),
		},
		.er = {
			.endCode = TO_U16(cause),
		},
	};
	
	if (!hotsyncPrvTx(&endReq, sizeof(endReq)))
		return 0;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return 0;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_SYNC_END) || rsp->err)
		return 0;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	
	return nArgs >= 0;
}

bool hotsyncDbDelete(uint8_t cardNo, const char *name)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsDbOpenOrDeleteReq dbo;
	} __attribute__((packed)) dbDeleteReq = {
		.hdr = {
			.pktTyp = HS_PKT_DB_DELETE,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsDbOpenOrDeleteReq),
		},
		.dbo = {
			.cardNo = cardNo,
			.openMode = 0,
		},
	};
	strcpy(dbDeleteReq.dbo.dbName, name);
	
	if (!hotsyncPrvTx(&dbDeleteReq, sizeof(dbDeleteReq)))
		return false;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return false;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_DB_DELETE) || rsp->err)
		return false;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);

	return nArgs >= 0;
}

enum HotsyncRecReadRet hotsyncRecRead(uint8_t dbHandle, uint16_t idx, uint16_t offset, uint16_t len, void *dst, uint32_t *recIdP, uint16_t *recSizeP, uint8_t *recAttrP, uint8_t *categoryP)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	const struct HsRecReadResp *rrsp;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsRecReadByIdxReq rr;
	} __attribute__((packed)) recReadReq = {
		.hdr = {
			.pktTyp = HS_PKT_REC_READ,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST + 1,
			.argLen = sizeof(struct HsRecReadByIdxReq),
		},
		.rr = {
			.dbHandle = dbHandle,
			.idx = TO_U16(idx),
			.ofst = TO_U16(offset),
			.len = TO_U16(len),
		},
	};
	
	if (!hotsyncPrvTx(&recReadReq, sizeof(recReadReq)))
		return HotsyncRecReadError;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return HotsyncRecReadError;
		
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_REC_READ))
		return HotsyncRecReadError;
	
	if (TO_U16(rsp->err) == HS_ERR_NOT_FOUND)		//no such record?
		return HotsyncRecDoesntExist;
	
	if (rsp->err)
		return HotsyncRecReadError;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsRecReadResp))
		return HotsyncRecReadError;
	
	rrsp = (const struct HsRecReadResp*)(mTxRxBuf + argsOfst[0]);
	if (recIdP)
		*recIdP = TO_U32(rrsp->recId);
	if (recSizeP)
		*recSizeP = TO_U16(rrsp->recSize);
	if (recAttrP)
		*recAttrP = rrsp->attr;
	if (categoryP)
		*categoryP = rrsp->category;
	
	memcpy(dst, rrsp->data, len);
	
	return HotsyncRecExists;
}

bool hotsyncRecWriteBegin(uint8_t dbHandle, uint16_t size, uint32_t recID, uint8_t recAttrs, uint8_t category)
{
	struct SlkHdr slkHdr = {.dstSlk = PKT_ADDR_DESKTOP_LINK, .srcSck = PKT_ADDR_DESKTOP_LINK, .pktTyp = PKT_TYP_HOTSYNC, .tid = hotsyncPrvGetTxTid()};
	struct HsDbCreateRsp *rspParam;
	struct SlkHdr hdr;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperSmall argWrapper;
		struct HsRecWriteReq wr;
	} __attribute__((packed)) recWriteReq = {
		.hdr = {
			.pktTyp = HS_PKT_REC_WRITE,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST | HS_HL_ARG_MASK_SMALL,
			.argLen = TO_U16(sizeof(struct HsRecWriteReq) + size),
		},
		.wr = {
			.dbHandle = dbHandle,
			.flags = 0x80,
			.recID = TO_U32(recID),
			.recAttrs = recAttrs,
			.category = category,
		},
	};
		
	return fragSendStartData(NULL, &mFs, &slkHdr, sizeof(recWriteReq) + size) && fragSendContinue(NULL, &mFs, &recWriteReq, sizeof(recWriteReq), NULL);
}

bool hotsyncRecWriteContinue(const void *data, uint16_t len)
{
	return fragSendContinue(NULL, &mFs, data, len, NULL);
}

bool hotsyncRecWriteFinish(uint32_t *recIDP)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4], rxLen;
	int_fast8_t nArgs;

	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return 0;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_REC_WRITE) || rsp->err)
		return 0;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsRecWriteResp))
		return false;
	
	if (recIDP)
		*recIDP = TO_U32(((const struct HsRecWriteResp*)(mTxRxBuf + argsOfst[0])) -> recID);
	
	return true;
}

bool hotsyncRecDelete(uint8_t dbHandle, uint32_t recID)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsRecDeleteReq dr;
	} __attribute__((packed)) recDeleteReq = {
		.hdr = {
			.pktTyp = HW_PKT_REC_DELETE,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsRecDeleteReq),
		},
		.dr = {
			.dbHandle = dbHandle,
			.recID = TO_U32(recID),
		},
	};
	
	if (!hotsyncPrvTx(&recDeleteReq, sizeof(recDeleteReq)))
		return false;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return false;
		
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HW_PKT_REC_DELETE) || rsp->err)
		return false;
	
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	
	return nArgs >= 0;
}

bool hotsyncInfoBlockRead(uint8_t dbHandle, bool appInfo, uint16_t offset, uint16_t len, void *dst, uint16_t *fullSizeP)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	const struct HsInfoBlockReadResp *rrsp;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsInfoBlockReadReq rr;
	} __attribute__((packed)) infoBlockReadReq = {
		.hdr = {
			.pktTyp = appInfo ? HS_PKT_APP_INFO_READ : HS_PKT_SORT_INFO_READ,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsInfoBlockReadReq),
		},
		.rr = {
			.dbHandle = dbHandle,
			.ofst = TO_U16(offset),
			.len = TO_U16(len),
		},
	};
	
	if (!hotsyncPrvTx(&infoBlockReadReq, sizeof(infoBlockReadReq)))
		return false;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return false;
		
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | infoBlockReadReq.hdr.pktTyp))
		return false;
	
	if (TO_U16(rsp->err) == HS_ERR_NOT_FOUND) {		//no such block?
		
		if (fullSizeP)
			*fullSizeP = 0;
		
		return true;
	}
	
	if (rsp->err)
		return false;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsInfoBlockReadResp))
		return false;
	
	rrsp = (const struct HsInfoBlockReadResp*)(mTxRxBuf + argsOfst[0]);
	if (fullSizeP)
		*fullSizeP = TO_U16(rrsp->fullSize);
	
	memcpy(dst, rrsp->data, len);
	
	return true;
}

bool hotsyncInfoBlockWriteBegin(uint8_t dbHandle, bool appInfo, uint16_t size)
{
	struct SlkHdr slkHdr = {.dstSlk = PKT_ADDR_DESKTOP_LINK, .srcSck = PKT_ADDR_DESKTOP_LINK, .pktTyp = PKT_TYP_HOTSYNC, .tid = hotsyncPrvGetTxTid()};
	struct HsDbCreateRsp *rspParam;
	struct SlkHdr hdr;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperSmall argWrapper;
		struct HsInfoBlockWriteReq wr;
	} __attribute__((packed)) infoBlockWriteReq = {
		.hdr = {
			.pktTyp = appInfo ? HS_PKT_APP_INFO_WRITE : HS_PKT_SORT_INFO_WRITE,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST | HS_HL_ARG_MASK_SMALL,
			.argLen = TO_U16(sizeof(struct HsInfoBlockWriteReq) + size),
		},
		.wr = {
			.dbHandle = dbHandle,
			.size = TO_U16(size),
		},
	};
	
	return fragSendStartData(NULL, &mFs, &slkHdr, sizeof(infoBlockWriteReq) + size) && fragSendContinue(NULL, &mFs, &infoBlockWriteReq, sizeof(infoBlockWriteReq), NULL);
}

bool hotsyncInfoBlockWriteContinue(const void *data, uint16_t len)
{
	return fragSendContinue(NULL, &mFs, data, len, NULL);
}

bool hotsyncInfoBlockWriteFinish(bool appInfo)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4], rxLen;
	int_fast8_t nArgs;

	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return 0;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | (appInfo ? HS_PKT_APP_INFO_WRITE : HS_PKT_SORT_INFO_WRITE)) || rsp->err)
		return 0;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	
	return nArgs >= 0;
}

int16_t hotsyncDbOpen(uint8_t cardNo, const char *name, uint8_t mode)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsDbOpenOrDeleteReq dbo;
	} __attribute__((packed)) dbOpenReq = {
		.hdr = {
			.pktTyp = HS_PKT_DB_OPEN,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsDbOpenOrDeleteReq),
		},
		.dbo = {
			.cardNo = cardNo,
			.openMode = mode,
		},
	};
	strcpy(dbOpenReq.dbo.dbName, name);
	
	if (!hotsyncPrvTx(&dbOpenReq, sizeof(dbOpenReq)))
		return -1;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return -1;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_DB_OPEN))
		return -1;
		
	if (rsp->err) {
		printf(" db open err 0x%04x\n", rsp->err);
		return -1;
	}
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < 1)
		return -1;

	return (uint16_t)*(uint8_t*)(mTxRxBuf + argsOfst[0]);
}

int16_t hotsyncDbCreate(struct HsDbCreateReq *req)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	struct HsDbCreateRsp *rspParam;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsDbCreateReq dbc;
	} __attribute__((packed)) dbCreateReq = {
		.hdr = {
			.pktTyp = HS_PKT_DB_CREATE,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsDbCreateReq),
		},
		.dbc = *req,
	};
	
	if (!hotsyncPrvTx(&dbCreateReq, sizeof(dbCreateReq)))
		return -1;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return -1;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_DB_CREATE))
		return -1;
	
	if (rsp->err) {
		printf(" db cr err 0x%04x\n", rsp->err);
		return -1;
	}
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsDbCreateRsp))
		return -1;
	
	rspParam = (struct HsDbCreateRsp*)(mTxRxBuf + argsOfst[0]);
	return (uint16_t)rspParam->dbHandle;
}

bool hotsyncDbClose(uint8_t dbHandle)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	struct HsDbCreateRsp *rspParam;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsCloseDbReq cdr;
		uint8_t padding;
	} __attribute__((packed)) dbCloseReq = {
		.hdr = {
			.pktTyp = HS_PKT_DB_CLOSE,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsCloseDbReq),
		},
		.cdr = {
			.dbHandle = dbHandle,
		},
	};
	
	if (!hotsyncPrvTx(&dbCloseReq, sizeof(dbCloseReq)))
		return false;

	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return false;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_DB_CLOSE) || rsp->err)
		return false;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	
	return nArgs >= 0;
}

bool hotsyncResDelete(uint8_t dbHandle, uint32_t resType, uint16_t resID)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsResDeleteReq dr;
	} __attribute__((packed)) resDeleteReq = {
		.hdr = {
			.pktTyp = HW_PKT_RES_DELETE,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsResDeleteReq),
		},
		.dr = {
			.dbHandle = dbHandle,
			.resType = TO_U32(resType),
			.resID = TO_U16(resID),
		},
	};
	
	if (!hotsyncPrvTx(&resDeleteReq, sizeof(resDeleteReq)))
		return false;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return false;
		
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HW_PKT_RES_DELETE) || rsp->err)
		return false;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	
	return nArgs >= 0;
}

bool hotsyncResRead(uint8_t dbHandle, uint16_t idx, uint16_t offset, uint16_t len, void *dst, uint32_t *resTypeP, uint16_t *resIDP, uint16_t *resSizeP)	//ask for too much, get garbge or error...
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	const struct HsResReadResp *rrsp;
	uint16_t argsOfst[4];
	int_fast8_t nArgs;
	struct SlkHdr hdr;
	uint16_t rxLen;
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsResReadByIdxReq rr;
	} __attribute__((packed)) infoBlockReadReq = {
		.hdr = {
			.pktTyp = HS_PKT_RES_READ,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsResReadByIdxReq),
		},
		.rr = {
			.dbHandle = dbHandle,
			.idx = TO_U16(idx),
			.ofst = TO_U16(offset),
			.len = TO_U16(len),
		},
	};
	
	if (!hotsyncPrvTx(&infoBlockReadReq, sizeof(infoBlockReadReq)))
		return false;
	
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen)) {
		printf("rx fail\n");
		return false;
	}
		
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_RES_READ)) {
		printf("reply type fail\n");
		return false;
	}
	
	if (TO_U16(rsp->err) == HS_ERR_NOT_FOUND) {		//no such resource?
		if (resSizeP)
			*resSizeP = 0;
		return true;
	}
	
	if (rsp->err) {
		printf("res read err 0x%04x\n", TO_U16(rsp->err));
		return false;
	}
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsResReadResp))
		return false;
	
	rrsp = (const struct HsResReadResp*)(mTxRxBuf + argsOfst[0]);
	if (resTypeP)
		*resTypeP = TO_U32(rrsp->resType);
	if (resIDP)
		*resIDP = TO_U16(rrsp->resID);
	if (resSizeP)
		*resSizeP = TO_U16(rrsp->resSize);
	
	memcpy(dst, rrsp->data, len);
	
	return true;
}

bool hotsyncResWriteBegin(uint8_t dbHandle, uint32_t resTyp, uint16_t resID, uint16_t len)
{
	struct SlkHdr slkHdr = {.dstSlk = PKT_ADDR_DESKTOP_LINK, .srcSck = PKT_ADDR_DESKTOP_LINK, .pktTyp = PKT_TYP_HOTSYNC, .tid = hotsyncPrvGetTxTid()};
	struct HsDbCreateRsp *rspParam;
	struct SlkHdr hdr;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperSmall argWrapper;
		struct HsResWriteReq rwr;
	} __attribute__((packed)) resCreateReq = {
		.hdr = {
			.pktTyp = HS_PKT_RES_WRITE,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST | HS_HL_ARG_MASK_SMALL,
			.argLen = TO_U16(sizeof(struct HsResWriteReq) + len),
		},
		.rwr = {
			.dbHandle = dbHandle,
			.resTyp = TO_U32(resTyp),
			.resID = TO_U16(resID),
			.resSize = TO_U16(len),
		},
	};
	
	return fragSendStartData(NULL, &mFs, &slkHdr, sizeof(resCreateReq) + len) && fragSendContinue(NULL, &mFs, &resCreateReq, sizeof(resCreateReq), NULL);
}

bool hotsyncResWriteContinue(const void *data, uint16_t len)
{
	return fragSendContinue(NULL, &mFs, data, len, NULL);
}

bool hotsyncResWriteFinish(void)
{
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t argsOfst[4], rxLen;
	int_fast8_t nArgs;

	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return 0;
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_RES_WRITE) || rsp->err)
		return 0;
	
	nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
	
	return nArgs >= 0;
}

//init is special: it does not use arg wrapping or HS_PKT_REPLY_MASK
uint32_t hotsyncInit(const void *incomingPkt, uint_fast16_t len, uint32_t desiredBaudrate)
{
	const struct HsInitPkt *wake = (struct HsInitPkt*)incomingPkt;
	uint32_t maxBaudrate, curBaudrate = 9600;
	struct HsInitPkt init = {
		.pktTyp = HS_PKT_INIT,
		.ver = {1, 0},
	};
	
	if (len < sizeof(struct HsInitPkt))
		return 0;
	
	if (wake->pktTyp != HS_PKT_WAKE || wake->ver[0] != 1 || wake->ver[1] > 3 || wake->sbz) {
		//fprintf(stderr, "wake packet inval %u %u %u\n", wake->ver[0], wake->ver[1], wake->sbz);
		return 0;
	}
	
	maxBaudrate = TO_U32(wake->maxBaud);
	//fprintf(stderr, "Got wake packet, max baudrate %u\n", (unsigned)maxBaudrate);
	
	if (maxBaudrate > curBaudrate && desiredBaudrate > curBaudrate) {
		
		init.flags |= HS_INIT_FLG_CHANGE_BAUDRATE;
		curBaudrate = desiredBaudrate > maxBaudrate ? maxBaudrate : desiredBaudrate;
	}
	init.maxBaud = TO_U32(curBaudrate);
	
	if (!hotsyncPrvTx(&init, sizeof(struct HsInitPkt)))
		return 0;
	
	//if baudrate changed, delay a bit, then change it
	if (init.flags & HS_INIT_FLG_CHANGE_BAUDRATE) {
		
		uint32_t start = slkExtGetCurTime();
		
		while (slkExtGetCurTime() - start < 10);	//100ms-ish
	}
	
	return curBaudrate;
}

bool hotsyncDbsEnum(uint8_t cardNo, bool includeROM, bool includeRAM, HotsyncDbsEnumF cbkF, void *cbkD)
{
	struct SlkHdr slkHdr = {.dstSlk = PKT_ADDR_DESKTOP_LINK, .srcSck = PKT_ADDR_DESKTOP_LINK, .pktTyp = PKT_TYP_HOTSYNC, .tid = hotsyncPrvGetTxTid()};
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	struct HsDbEnumResult *result;
	uint16_t argsOfst[4], rxLen;
	struct SlkHdr hdr;
	int_fast8_t nArgs;
	
	struct {
		struct HsHlReqHdr hdr;
		struct HsHlArgWrapperTiny argWrapper;
		struct HsEnumDbsReq enumReq;
	} __attribute__((packed)) dbEnumReq = {
		.hdr = {
			.pktTyp = HS_PKT_GET_DB_LIST,
			.nArgs = 1,
		},
		.argWrapper = {
			.argIdAndLenType = HS_HL_ARG_IDX_FIRST,
			.argLen = sizeof(struct HsEnumDbsReq),
		},
		.enumReq = {
			.flags = (includeROM ? 0x40 : 0) | (includeRAM ? 0x80 : 0),
			.cardNo = cardNo,
			.startIdx = 0,
		},
	};
	
	while (1) {
	
		if (!hotsyncPrvTx(&dbEnumReq, sizeof(dbEnumReq)))
			return false;
		
		rxLen = HOTSYNC_MIN_BUF_SIZE;
		if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
			return false;
	
		if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_GET_DB_LIST))
			return false;
		
		if (TO_U16(rsp->err) == HS_ERR_NOT_FOUND)
			break;
		
		if (rsp->err)
			return false;
		
		nArgs = hotsyncPrvArgsParse(mTxRxBuf, rxLen, sizeof(argsOfst) / sizeof(*argsOfst) - 1, argsOfst);
		if (nArgs < 1 || argsOfst[1] - argsOfst[0] < sizeof(struct HsDbEnumResult))
			return false;
	
		result = (struct HsDbEnumResult*)(mTxRxBuf + argsOfst[0]);
		
		if (!result->numItems)
			break;
		
		if (result->dbi.structSz <= sizeof(struct HsDbInfo))
			return false;
		
		if (!cbkF(cbkD, &result->dbi))
			break;
		
		if (!(result->enumFlags & 0x80))	//any more?
			break;
		
		//make sure we do not overflow on accident
		if (!(dbEnumReq.enumReq.startIdx = TO_U16(TO_U16(result->lastIndex) + 1)))
			break;
	}
	
	return true;
}

bool hotsyncRPC(struct HsRpcPkt *pkt, uint16_t len)
{
	/*
		due to bugs in at least palmOS 1.0, we cnanot send this one chunked. This causes palmos to ue a MemHandle to keep track of the state
		and it does not then properly use that to perform the RPC causing it to try to execute whatever is at address 0x000002 as RPC...
		Thus our HsRpcPkt struct includes the header in it :)
	*/
	struct SlkHdr slkHdr = {.dstSlk = PKT_ADDR_DESKTOP_LINK, .srcSck = PKT_ADDR_DESKTOP_LINK, .pktTyp = PKT_TYP_HOTSYNC, .tid = hotsyncPrvGetTxTid()};
	const struct HsHlRspHdr *rsp = (const struct HsHlRspHdr*)mTxRxBuf;
	uint16_t rxLen;
	
	
	if (len < sizeof(struct HsHlReqHdr) || (len & 1))
		return false;
	
	if (!hotsyncPrvTx(pkt, len))
		return 0;
		
	rxLen = HOTSYNC_MIN_BUF_SIZE;
	if (!hotsyncPrvRx(mTxRxBuf, &rxLen))
		return false;
	
	
	if (rsp->pktTyp != (HS_PKT_REPLY_MASK | HS_PKT_RPC) || rsp->err)
		return false;
	
	if (rxLen - sizeof(struct HsHlRspHdr) != len - sizeof(struct HsHlReqHdr))
		return false;
	
	//skip 
	memcpy(((char*)pkt) + sizeof(struct HsHlReqHdr), rsp + 1, len - sizeof(struct HsHlReqHdr));
	return true;
}

bool hotsyncRpcFtrGet(uint32_t ftrCrid, uint16_t ftrId, uint32_t *ftrValP, uint16_t *errP)
{
	struct {
		struct HsRpcPkt rpc;
		struct HsRpcParam p3;
		uint32_t ftrVal;
		struct HsRpcParam p2;
		uint16_t ftrID;
		struct HsRpcParam p1;
		uint32_t ftrType;
	} __attribute__((packed)) ftrGet = {
		.rpc = {
			.val_0x2d = 0x2d,
			.trapNo = TO_U16(0xA27B),	//FtrGet
			.nParams = TO_U16(3),
		},
		.p3 = {
			.boolPassByRef = true,
			.size = 4,
		},
		.p2 = {
			.size = 2,
		},
		.ftrID = TO_U16(ftrId),
		.p1 = {
			.size = 4,
		},
		.ftrType = TO_U32(ftrCrid),
	};
	
	if (!hotsyncRPC(&ftrGet.rpc, sizeof(ftrGet)))
		return false;
	
	if (errP)
		*errP = TO_U32(ftrGet.rpc.retD0);
	
	if (ftrValP)
		*ftrValP = TO_U32(ftrGet.ftrVal);
	
	return true;
}

static bool hotsyncRpcPrefGetSetAppPreferencesV10(uint32_t crid, int_fast16_t version, void *data, uint_fast16_t dataSz, bool set, bool *boolOutP)
{
	uint_fast16_t roundedDataSz = (dataSz + 1) / 2 * 2;
	struct HsUnalignedU16 *v16;
	struct HsUnalignedU32 *v32;
	struct HsRpcParam *param;
	struct HsRpcPkt *pkt;
	char *dataPtr;
	uint32_t sz;
	
	pkt = (struct HsRpcPkt*)alloca(sz = sizeof(struct HsRpcPkt) + sizeof(struct HsRpcParam) * 4 + sizeof(uint32_t) + sizeof(uint16_t) + roundedDataSz + sizeof(uint16_t));
	memset(pkt, 0, sz);
	
	pkt->val_0x2d = 0x2d;
	pkt->trapNo = TO_U16(set ? 0xA22F : 0xA22E);	//PrefSetAppPreferencesV10 : PrefGetAppPreferencesV10
	pkt->nParams = TO_U16(4);
	param = (struct HsRpcParam*)(pkt + 1);
	param->size = sizeof(uint16_t);
	v16 = (struct HsUnalignedU16*)(param + 1);
	v16->val = TO_U16(dataSz);
	param = (struct HsRpcParam*)(v16 + 1);
	param->boolPassByRef = true;
	param->size = roundedDataSz;
	dataPtr = (char*)(param + 1);
	if (set)
		memcpy(dataPtr, data, dataSz);
	param = (struct HsRpcParam*)(dataPtr + roundedDataSz);
	param->size = sizeof(int16_t);
	v16 = (struct HsUnalignedU16*)(param + 1);
	v16->val = TO_U16(version);
	param = (struct HsRpcParam*)(v16 + 1);
	param->size = sizeof(uint32_t);
	v32 = (struct HsUnalignedU32*)(param + 1);
	v32->val = TO_U32(crid);
	
	if (!hotsyncRPC(pkt, sz))
		return false;
	
	if (!set)
		memcpy(data, dataPtr, dataSz);
	
	if (boolOutP)
		*boolOutP = !!pkt->retD0;
}

bool hotsyncRpcPrefSetAppPreferencesV10(uint32_t crid, int_fast16_t version, const void *data, uint_fast16_t dataSz)
{
	return hotsyncRpcPrefGetSetAppPreferencesV10(crid, version, (void *)data, dataSz, true, NULL);
}

bool hotsyncRpcPrefGetAppPreferencesV10(uint32_t crid, int_fast16_t version, void *data, uint_fast16_t dataSz, bool *retValP)
{
	return hotsyncRpcPrefGetSetAppPreferencesV10(crid, version, data, dataSz, false, retValP);
}

int16_t hotsyncRpcMemReadByte(uint32_t addr)
{
	struct {
		struct HsRpcPkt rpc;
		struct HsRpcParam p3;
		uint32_t length;
		struct HsRpcParam p2;
		uint32_t srcAddr;
		struct HsRpcParam p1;
		uint8_t byteVal;
		uint8_t padding;
	} __attribute__((packed)) memRead = {
		.rpc = {
			.val_0x2d = 0x2d,
			.trapNo = TO_U16(0xA026),	//MemMove
			.nParams = TO_U16(3),
		},
		.p3 = {
			.size = 4,
		},
		.length = TO_U32(1),
		.p2 = {
			.size = 4,
		},
		.srcAddr = TO_U32(addr),
		.p1 = {
			.boolPassByRef = true,
			.size = 1,
		},
	};
	
	if (!hotsyncRPC(&memRead.rpc, sizeof(memRead)))
		return -1;
		
	return (uint16_t)(uint8_t)memRead.byteVal;
}

