#include <stdbool.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include "timebase.h"
#include "printf.h"
#include "pinmux.h"
#include "lpc802.h"
#include "sd.h"


#define VERBOSE				0


#define FLAG_TIMEOUT		0x80
#define FLAG_PARAM_ERR		0x40
#define FLAG_ADDR_ERR		0x20
#define FLAG_ERZ_SEQ_ERR	0x10
#define FLAG_CMD_CRC_ERR	0x08
#define FLAG_ILLEGAL_CMD	0x04
#define FLAG_ERZ_RST		0x02
#define FLAG_IN_IDLE_MODE	0x01



struct SD {
	uint32_t numSec;
	union SdFlags flags;
	uint8_t mid;
	uint16_t oid;
	uint32_t snum;
	
	uint32_t numSecLeft;		//for long operation
};



static struct SD mSD;



static void sdSpiByteTxOnly(uint_fast8_t val)
{
	while (!(SPI0->STAT & 2));
	SPI0->TXDATCTL = 0x07500000 | val;
}

static uint8_t sdSpiByte(uint_fast8_t val)
{
	while (!(SPI0->STAT & 2));
	SPI0->TXDATCTL = 0x07100000 | val;
	while (!(SPI0->STAT & 1));
	val = SPI0->RXDAT;
	
	return val;
}

static void sdSpiBytePipelinedStart(uint_fast8_t val)
{
	while (!(SPI0->STAT & 2));
	SPI0->TXDATCTL = 0x07000000 | val;
}

static uint8_t sdSpiBytePipelinedCont(uint_fast8_t val)
{
	while (!(SPI0->STAT & 2));
	SPI0->TXDAT = val;
	while (!(SPI0->STAT & 1));
	val = SPI0->RXDAT;
	
	return val;
}

static uint8_t sdSpiBytePipelinedEnd(void)
{
	uint8_t val;
	
	while (!(SPI0->STAT & 1));
	val = SPI0->RXDAT;
	SPI0->STAT = 0x80;
	
	return val;
}

static void sdChipSelect(void)
{
	GPIO->CLR[0] = 1 << PIN_NO_SD_nCS;
}

static void sdChipDeselect(void)
{
	while (!(SPI0->STAT & 0x100));
	GPIO->SET[0] = 1 << PIN_NO_SD_nCS;
	sdSpiByteTxOnly(0xff);
	while (!(SPI0->STAT & 0x100));
}

#ifdef ENABLE_PRINTF_LOGGING	//space as at a premium

static uint_fast8_t sdCrc7(uint8_t* chr, uint_fast8_t cnt, uint_fast8_t crc)
{
	uint_fast8_t i, a, data;
	
	for (a = 0; a < cnt; a++) {
	
		data = chr[a];
		
		for (i = 0; i < 8; i++, data <<= 1) {
		
			crc <<= 1;
			
			if ((data & 0x80) ^ (crc & 0x80))
				crc ^= 0x09;
		}
	}
	return crc;
}

#else // be fast

static uint_fast8_t sdCrc7(uint8_t* chr, uint_fast8_t cnt, uint_fast8_t crc)
{
	static const uint8_t mCrcTab7[] = {		//generated from the iterative func :)
		0x00, 0x09, 0x12, 0x1b, 0x24, 0x2d, 0x36, 0x3f, 0x48, 0x41, 0x5a, 0x53, 0x6c, 0x65, 0x7e, 0x77,
		0x19, 0x10, 0x0b, 0x02, 0x3d, 0x34, 0x2f, 0x26, 0x51, 0x58, 0x43, 0x4a, 0x75, 0x7c, 0x67, 0x6e,
		0x32, 0x3b, 0x20, 0x29, 0x16, 0x1f, 0x04, 0x0d, 0x7a, 0x73, 0x68, 0x61, 0x5e, 0x57, 0x4c, 0x45,
		0x2b, 0x22, 0x39, 0x30, 0x0f, 0x06, 0x1d, 0x14, 0x63, 0x6a, 0x71, 0x78, 0x47, 0x4e, 0x55, 0x5c,
		0x64, 0x6d, 0x76, 0x7f, 0x40, 0x49, 0x52, 0x5b, 0x2c, 0x25, 0x3e, 0x37, 0x08, 0x01, 0x1a, 0x13,
		0x7d, 0x74, 0x6f, 0x66, 0x59, 0x50, 0x4b, 0x42, 0x35, 0x3c, 0x27, 0x2e, 0x11, 0x18, 0x03, 0x0a,
		0x56, 0x5f, 0x44, 0x4d, 0x72, 0x7b, 0x60, 0x69, 0x1e, 0x17, 0x0c, 0x05, 0x3a, 0x33, 0x28, 0x21,
		0x4f, 0x46, 0x5d, 0x54, 0x6b, 0x62, 0x79, 0x70, 0x07, 0x0e, 0x15, 0x1c, 0x23, 0x2a, 0x31, 0x38,
		0x41, 0x48, 0x53, 0x5a, 0x65, 0x6c, 0x77, 0x7e, 0x09, 0x00, 0x1b, 0x12, 0x2d, 0x24, 0x3f, 0x36,
		0x58, 0x51, 0x4a, 0x43, 0x7c, 0x75, 0x6e, 0x67, 0x10, 0x19, 0x02, 0x0b, 0x34, 0x3d, 0x26, 0x2f,
		0x73, 0x7a, 0x61, 0x68, 0x57, 0x5e, 0x45, 0x4c, 0x3b, 0x32, 0x29, 0x20, 0x1f, 0x16, 0x0d, 0x04,
		0x6a, 0x63, 0x78, 0x71, 0x4e, 0x47, 0x5c, 0x55, 0x22, 0x2b, 0x30, 0x39, 0x06, 0x0f, 0x14, 0x1d,
		0x25, 0x2c, 0x37, 0x3e, 0x01, 0x08, 0x13, 0x1a, 0x6d, 0x64, 0x7f, 0x76, 0x49, 0x40, 0x5b, 0x52,
		0x3c, 0x35, 0x2e, 0x27, 0x18, 0x11, 0x0a, 0x03, 0x74, 0x7d, 0x66, 0x6f, 0x50, 0x59, 0x42, 0x4b,
		0x17, 0x1e, 0x05, 0x0c, 0x33, 0x3a, 0x21, 0x28, 0x5f, 0x56, 0x4d, 0x44, 0x7b, 0x72, 0x69, 0x60,
		0x0e, 0x07, 0x1c, 0x15, 0x2a, 0x23, 0x38, 0x31, 0x46, 0x4f, 0x54, 0x5d, 0x62, 0x6b, 0x70, 0x79,
	};
	
	uint_fast8_t i;
	
	for (i = 0; i < cnt; i++)
		crc = mCrcTab7[crc * 2 + *chr++];
	
	return crc;
}
#endif

static void sdPrvSendCmd(uint8_t cmd, uint32_t param, bool crc)
{
	sdSpiByteTxOnly(0x40 | cmd);
	sdSpiByteTxOnly(param >> 24);
	sdSpiByteTxOnly(param >> 16);
	sdSpiByteTxOnly(param >> 8);
	sdSpiByteTxOnly(param);
	
	if (crc) {
		
		uint8_t send[] = {0x40 | cmd, param >> 24, param >> 16, param >> 8, param};
	
		sdSpiByteTxOnly(sdCrc7(send, sizeof(send), 0) * 2 + 1);
	}
}

static uint_fast8_t sdPrvSimpleCommand(uint_fast8_t cmd, uint32_t param, bool crc, bool end_cs)	//do a command, return R1 reply
{
	uint_fast8_t ret, i = 0;
	
	sdChipSelect();
	
	sdPrvSendCmd(cmd, param, crc);
	
	do {		//our max wait time is 128 byte clocks (1024 clock ticks)
		
		ret = sdSpiByte(0xFF);
		
	} while (i++ < 128 && (ret == 0xFF));
	
	if (end_cs)
		sdChipDeselect();
	
	if (VERBOSE)
		printf("cmd %2u (0x%08x) -> 0x%02x\n", cmd, param, ret);
	
	return ret;
}

static uint_fast8_t sdPrvACMD(uint_fast8_t cmd, uint32_t param, bool crc, bool end_cs)
{
	uint_fast8_t ret;
	
	ret = sdPrvSimpleCommand(55, 0, crc, true);
	if (ret & (FLAG_TIMEOUT | FLAG_ILLEGAL_CMD))
		return ret;
	
	return sdPrvSimpleCommand(cmd, param, crc, end_cs);
}

static bool sdPrvCardInit(bool isSdCard, bool signalHighCapacityHost)
{
	uint32_t param, time = 0, resp;
	bool first = true;
	
	param = (signalHighCapacityHost ? 0x40000000 : 0x00000000);
	
	while (time++ < 1024) {
	
		//busy bit is top bit of OCR, which is top bit of resp[1]
		//busy bit at 1 means init complete
		//see pages 85 and 26
		
		resp = isSdCard ? sdPrvACMD(41, param, true, true) : sdPrvSimpleCommand(1, param, true, true);
		
		if (resp & FLAG_TIMEOUT)
			break;
		
		if (!first && !(resp & FLAG_IN_IDLE_MODE))
			return true;
		first = false;
	}
	
	return false;
}

static uint_fast8_t __attribute__((noinline)) sdPrvReadData(uint8_t* data, uint_fast16_t sz)	//length must be divisible by 8, pointer must be halfword aligned
{
	uint_fast16_t tries = 1000000;
	
	do {
		uint_fast8_t ret;
	
		ret = sdSpiByte(0xFF);
		if ((ret & 0xF0) == 0x00) {
			return ret;	//fail
		}
		if (ret == 0xFE)
			break;
		tries--;
	} while(tries);
	
	if (!tries) {
		return 0xFF;
	}
	if (((((uintptr_t)data) | sz) & 1) || SPI0->DIV) {	//slow case
		
		while (sz--)
			*data++ = sdSpiByte(0xFF);
		sdSpiByte(0xFF);
		sdSpiByte(0xFF);
	
	}
	else {								//fast case

		uint32_t dummy;
		
		while (!(SPI0->STAT & 2));
		SPI0->TXDATCTL = 0x0f00ffff;	//done in 16-bit mode for speed
		sz /= 8;
		
		asm volatile(
			
			//we wait once... then we rely on cycle counting
			//which is idiotic since accesses to the spi peripheral take more cycles than expected
			//an unerrored APB read/write takes 4 cy if no wait states are needed.
			"1:							\n\t"
			"	ldr  %0, [%3, %4]		\n\t"
			"	lsl  %0, #31			\n\t"	//TXRDY->C, RXRDY->N
			"	bpl  1b					\n\t"	//no rx data
			
			"3:							\n\t"	//once we sync, we are on every 16 cy, no need to wait
			
			"	str  %7, [%3, %5]		\n\t"	//6 cy (4 cycles on the apb + 2 in the cpu)
			"	ldr  %0, [%3, %6]		\n\t"	//6 cy (4 cycles on the apb + 2 in the cpu)
			
			"	rev16 %0, %0			\n\t"	//1 cy
			"	strh %0, [%1]			\n\t"	//2 cy
	
			//we expect to need nops here but we do not. do not ask
	
			"	str  %7, [%3, %5]		\n\t"	//6 cy (4 cycles on the apb + 2 in the cpu)
			"	ldr  %0, [%3, %6]		\n\t"	//6 cy (4 cycles on the apb + 2 in the cpu)
			
			"	rev16 %0, %0			\n\t"	//1 cy
			"	strh %0, [%1, #2]		\n\t"	//2 cy
	
			//we expect to need nops here but we do not. do not ask
	
			"	str  %7, [%3, %5]		\n\t"	//6 cy (4 cycles on the apb + 2 in the cpu)
			"	ldr  %0, [%3, %6]		\n\t"	//6 cy (4 cycles on the apb + 2 in the cpu)
			
			"	rev16 %0, %0			\n\t"	//1 cy
			"	strh %0, [%1, #4]		\n\t"	//2 cy
			"	sub  %2, #1				\n\t"	//1 cy	(loop maintenance)
			
			//we expect to need nops here but we do not. do not ask
	
			"	str  %7, [%3, %5]		\n\t"	//6 cy (4 cycles on the apb + 2 in the cpu)
			"	ldr  %0, [%3, %6]		\n\t"	//6 cy (4 cycles on the apb + 2 in the cpu)
			
			"	rev16 %0, %0			\n\t"	//1 cy
			"	strh %0, [%1, #6]		\n\t"	//2 cy
			"	add  %1, %8				\n\t"	//1 cy		high reg used to not clobber flags
	
			//we expect to need nops here but we do not. do not ask
	
			"	bne  3b					\n\t"	//2 cy if taken
			:"=&r"(dummy), "+r"(data), "+r"(sz)
			:"r"(SPI0_BASE), "i"(offsetof(SPI_Type, STAT)), "i"(offsetof(SPI_Type, TXDAT)), "i"(offsetof(SPI_Type, RXDAT)), "r"(0xffff), "h"(8)
			:"memory", "cc"
		);
		
		while (!(SPI0->STAT & 1));
		(void)SPI0->RXDAT;
		SPI0->STAT = 0x80;	
	}
	
	return 0;
}

static uint8_t sdPrvWriteData(const uint8_t *data, uint_fast16_t sz, bool isMultiblock)
{
	uint_fast16_t tries = 10000;
	uint_fast8_t ret;
	
	sdSpiByteTxOnly(isMultiblock ? 0xFC : 0xFE);	//start block
	
	while (sz--)
		sdSpiByteTxOnly(*data++);
	
	//crc
	sdSpiByteTxOnly(0);
	sdSpiByteTxOnly(0);
	
	//wait for a reply
	do {
		ret = sdSpiByte(0xFF);
		if ((ret & 0x11) == 0x01)
			return ret;
		tries--;
	} while(tries);
	
	return 0xFF;
}

static bool sdPrvPrgBusyWait(void)
{
	uint_fast16_t tries = 100000;
	
	while (sdSpiByte(0xFF) != 0xFF) {
		
		if (!--tries)
			return false;	
	}
	
	return true;
}

static uint32_t sdPrvGetBits(const uint8_t* data, uint32_t numBytesInArray, int32_t startBit, uint32_t len)	//for CID and CSD data..
{
	uint32_t ret = 0;
	int32_t i;
	
	for (i = startBit + len - 1; i >= startBit; i--)
		ret = (ret << 1) + ((data[numBytesInArray - 1 - i / 8] >> (i % 8)) & 1);
	
	return ret;
}

bool sdCardInit(void)
{
	uint_fast8_t i, v, j;
	uint8_t respBuf[64];
	uint32_t ocr = 0;
	uint64_t t;
	
	memset(&mSD, 0, sizeof(mSD));
	
	for (i = 0; i < 16; i++)	//give card time to init with CS deasserted
		sdSpiByteTxOnly(0xff);
	
	timebaseDelay(CLOCKRATE / 1000);
		
	for (i = 0; i < 16; i++) {	//try CMD0 a few times
		
		v = sdPrvSimpleCommand(0, 0, true, true);
		if (v == FLAG_IN_IDLE_MODE)
			break;
	}
	if (v & FLAG_TIMEOUT)
		return false;
	if (!(v & FLAG_IN_IDLE_MODE))
		return false;
	
	v = sdPrvSimpleCommand(8, 0x000001AAUL, true, false);	//try CMD8 to init SDHC cards
	if (v & FLAG_TIMEOUT) {
		sdChipDeselect();
		return false;
	}
	mSD.flags.v2 = !(v & FLAG_ILLEGAL_CMD);
	if (mSD.flags.v2) {
		//ignore the reply
		sdSpiByteTxOnly(0xFF);
		sdSpiByteTxOnly(0xFF);
		sdSpiByteTxOnly(0xFF);
		sdSpiByteTxOnly(0xFF);
	}
	sdChipDeselect();
	
	v = sdPrvSimpleCommand(55, 0, true, true);			//see if this is SD or MMC
	if (v & FLAG_TIMEOUT)
		return false;
	mSD.flags.SD = !(v & FLAG_ILLEGAL_CMD);
	
	if (mSD.flags.SD) {

		if (!sdPrvCardInit(true, true) && !sdPrvCardInit(true, false))
			return false;
	}
	else{
	
		if (!sdPrvCardInit(false, true) && !sdPrvCardInit(false, false))
			return false;
	}
	
	//check for HC
	v = sdPrvSimpleCommand(58, 0, false, false);
	if (v)
		return false;
	
	//read 32 bits of ocr
	for (v = 0; v < 4; v++)
		ocr = (ocr << 8) + sdSpiByte(0xff);
	sdChipDeselect();
	mSD.flags.HC = !!(ocr & 0x40000000);
	
	v = sdPrvSimpleCommand(59, 0, true, true);		//crc off
	if (v & FLAG_TIMEOUT)
		return false;

	v = sdPrvSimpleCommand(9, 0, false, false);		//read CSD
	if (v & FLAG_TIMEOUT)
		return false;
	
	v = sdPrvReadData(respBuf, 16);
	sdChipDeselect();
	if (v)
		return false;
	
	if (VERBOSE) {
		printf("CSD:");
		for (i = 0; i < 16; i++)
			printf(" %02x", respBuf[i]);
		printf("\n\n");
	}
	
	if (VERBOSE)
		printf("CSD ver %u\n", sdPrvGetBits(respBuf, 16, 126, 2) + 1);

	//calc capacity
	if (mSD.flags.SD && sdPrvGetBits(respBuf, 16, 126, 2)) {	//v2 SD csd
		
		mSD.numSec = 1024 * (1 + sdPrvGetBits(respBuf, 16, 48, 22));
		
		if (VERBOSE)
			printf("c_size: 0x%08x\n", sdPrvGetBits(respBuf, 16, 48, 22));
	}
	else {
		
		uint32_t cSize = sdPrvGetBits(respBuf, 16, 62, 12);
		uint32_t cSizeMult = sdPrvGetBits(respBuf, 16, 47, 3);
		uint32_t readBlLen = sdPrvGetBits(respBuf, 16, 80, 4);
		uint32_t shiftBy = readBlLen + cSizeMult + 2;
		
		if (!mSD.flags.SD && sdPrvGetBits(respBuf, 16, 126, 2) >= 2 && sdPrvGetBits(respBuf, 16, 122, 4) >= 4) {
			
			if (VERBOSE)
				printf("MMC+ over 4GB not supported due to lack of testing\n");
			
			return false;
		}
		
		if (shiftBy < 9) {	//fatal
			if (VERBOSE)
				printf("card shift size invalid\n");
			return false;
		}
		shiftBy -= 9;
		
		mSD.numSec = (cSize + 1) << shiftBy;
	}
	
	mSD.flags.RO = false;
	
	if (sdPrvGetBits(respBuf, 16, 12, 1)) {
		
		if (VERBOSE)
			printf("Card temporarily write protected\n");
		mSD.flags.RO = true;
	}
	
	if (sdPrvGetBits(respBuf, 16, 13, 1)) {
		
		if (VERBOSE)
			printf("Card permanently write protected\n");
		mSD.flags.RO = true;
	}
	
	if (VERBOSE)
		printf("CARD capacity: %u 512-byte blocks\n", mSD.numSec);
	
	if (mSD.flags.SD) {		//some SD-specific things
		
		v = sdPrvACMD(51, 0, false, false);
		if (v & FLAG_TIMEOUT)
			return false;
		
		v = sdPrvReadData(respBuf, 8);
		sdChipDeselect();
		if (v)
			return false;
		
		if (VERBOSE) {
			printf("SCR:");
			for (i = 0; i < 8; i++)
				printf(" %02x", respBuf[i]);
			printf("\n\n");
		}
		
		if (!(respBuf[1] & 1)) {
			if (VERBOSE)
				printf("card claims to not support 1-bit mode...clearly SCR is wrong\n");
			return false;
		}
		
		mSD.flags.cmd23supported = !!(respBuf[3] & 2);
		
		v = sdPrvACMD(13, 0, false, false);
		if (v & FLAG_TIMEOUT)
			return false;
		
		//that command sends a 2-byte reply, so consume the second byte
		(void)sdSpiByte(0xff);
		
		v = sdPrvReadData(respBuf, 64);
		sdChipDeselect();
		if (v)
			return false;
		
		if (VERBOSE) {
			printf("SD_STATUS:");
			for (i = 0; i < 64; i++)
				printf(" %02x", respBuf[i]);
			printf("\n\n");
		}
		
		mSD.flags.hasDiscard = !!(respBuf[24] & 2);
	}
	
	//speed the clock up
	SPI0->DIV = 0;
	
	//set block size
	v = sdPrvSimpleCommand(16, SD_BLOCK_SIZE, true, true);		//crc off
	if (v & FLAG_TIMEOUT)
		return false;
	
	v = sdPrvSimpleCommand(10, 0, false, false);		//read CID
	if (v & FLAG_TIMEOUT)
		return false;
	v = sdPrvReadData(respBuf, 16);
	sdChipDeselect();
	if (v)
		return false;
	
	//parse CID
	mSD.mid = sdPrvGetBits(respBuf, 16, 120, 8);
	mSD.oid = sdPrvGetBits(respBuf, 16, 104, mSD.flags.SD ? 16 : 8);
	mSD.snum = sdPrvGetBits(respBuf, 16, mSD.flags.SD ? 24 : 16, 32);

	mSD.flags.inited = true;

	return true;
}

uint32_t sdGetNumSecs(void)
{
	return mSD.numSec;
}

bool sdSecRead(uint32_t sec, uint8_t *dst)
{
	uint_fast8_t v;
	
	#ifdef ENABLE_ACTIVITY_LED
		GPIO->CLR[0] = 1 << PIN_NO_LED;
	#endif
	
	v = sdPrvSimpleCommand(17, mSD.flags.HC ? sec : sec << 9, false, false);
	if (v & FLAG_TIMEOUT) {
		if (VERBOSE)
			printf("command timeout\n");
		#ifdef ENABLE_ACTIVITY_LED
			GPIO->SET[0] = 1 << PIN_NO_LED;
		#endif
		return false;
	}
	
	v = sdPrvReadData(dst, SD_BLOCK_SIZE);
	sdChipDeselect();
	
	#ifdef ENABLE_ACTIVITY_LED
		GPIO->SET[0] = 1 << PIN_NO_LED;
	#endif
	
	return !v;
}

bool sdSecWrite(uint32_t sec, const uint8_t *src)
{
	bool busyDone;
	uint_fast8_t v;
	
	#ifdef ENABLE_ACTIVITY_LED
		GPIO->CLR[0] = 1 << PIN_NO_LED;
	#endif
	
	v = sdPrvSimpleCommand(24, mSD.flags.HC ? sec : sec << 9, false, false);
	if (v & FLAG_TIMEOUT) {
		#ifdef ENABLE_ACTIVITY_LED
			GPIO->SET[0] = 1 << PIN_NO_LED;
		#endif
		return false;
	}
	
	v = sdPrvWriteData(src, SD_BLOCK_SIZE, false);
	busyDone = sdPrvPrgBusyWait();
	
	sdChipDeselect();
	
	#ifdef ENABLE_ACTIVITY_LED
		GPIO->SET[0] = 1 << PIN_NO_LED;
	#endif
	
	return (v & 0x1f) == 0x05 && busyDone;
}

void sdGetInfo(uint8_t *midP, uint16_t *oidP, uint32_t *snumP)
{
	if (midP)
		*midP = mSD.mid;
	
	if (oidP)
		*oidP = mSD.oid;
	
	if (snumP)
		*snumP = mSD.snum;
}

uint8_t sdGetFlags(void)
{
	return mSD.flags.value;
}

void sdHwInit(void)
{
	//SPI nCS value
	GPIO->SET[0] = 1 << PIN_NO_SD_nCS;

	//spi 
	SYSCON->SPI0CLKSEL = 0;
	IOCON->PIO[mIoconRegOrder[PIN_NO_SD_MISO]] = 0x30;	//pull up
	IOCON->PIO[mIoconRegOrder[PIN_NO_SD_MOSI]] = 0x20;
	IOCON->PIO[mIoconRegOrder[PIN_NO_SD_SCLK]] = 0x20;
	IOCON->PIO[mIoconRegOrder[PIN_NO_SD_nCS]] = 0x20;
	GPIO->DIRCLR[0] = (1 << PIN_NO_SD_MISO);
	GPIO->DIRSET[0] = (1 << PIN_NO_SD_MOSI) | (1 << PIN_NO_SD_SCLK) | (1 << PIN_NO_SD_nCS);
	SWM0->PINASSIGN.PINASSIGN2 = (SWM0->PINASSIGN.PINASSIGN2 &~ 0x00ffffff) | ((PIN_NO_SD_SCLK) << 0) | ((PIN_NO_SD_MOSI) << 8) | ((PIN_NO_SD_MISO) << 16);
	SPI0->CFG = 0;	//off
	SPI0->DIV = (CLOCKRATE + 399999) / 400000;
	SPI0->TXCTL = 0x07000000;	//frames are bytes
	SPI0->CFG = 0b110101;	//on, mode3
	SPI0->DLY = 0;
}






